""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """

import time
import re
import ipaddress
import serial
from itertools import tee, islice, chain
from datetime import datetime


from common import utility, logger
from common import gvars
from common import response ,dutpipe



SUCCESS = gvars.SUCCESS
FAILURE = gvars.FAILURE
addba_reject_group = ["HE-5.16.1","HE-5.22.1","HE-5.23.1","HE-5.52.1","HE-5.59.1"]


def isExist(key):
    if key in gvars.SIGMA_TEST_DATA:
        return True
    return False

def _parse_from_log(log, label):
    resp = response.Response()
    # logger.INFO("Looking for {} from the following logs : \n {}".format(label,log))
    if label in log:
        for line in log.split("\n"):
            if label in line:
                resp.data = ":".join(line.split(":")[1:]).strip()
                resp.status = SUCCESS
                return resp

    else:
        resp.status = FAILURE
        resp.data = "Unable to find '%s' in log text '%s'" % (label, log)
        return resp

def _parse_ip_from_log(log):
    resp = response.Response()
    if gvars.VERBOSE:
        logger.INFO("Looking for ip from the following logs : \n {}".format(log))
    if "ip:" in log:
        for line in log.split("\n"):
            if "ip:" in line:
                data = ":".join(line.split(":")[1:]).strip()
                try:
                    _ = ipaddress.ip_address(data)
                except:
                    continue
                resp.data = data
                resp.status = SUCCESS
                return resp
    resp.status = FAILURE
    resp.data = "Unable to find in log text '%s'" % (log)
    return resp



def get_params_attribute(params, key):
    params = params.split(",")
    for i in range(len(params)):
        if params[i] == key:
            return params[i + 1]
    return None

def fetch_device_info():
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "version")
    return _parse_from_log(log, 'IDF Version')

def sta_set_pwrsave(mode):
    resp  = response.Response()
    if mode.lower() == "on":
        cli = "power --state 1"
    elif mode.lower() == "off":
        cli = "power --state 0"
    else:
        cli = "power --state 0"

    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
    resp.status = SUCCESS
    return resp

def get_card_type():
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "version")
    return _parse_from_log(log, 'model')

def reset_dut(prog):
    resp = response.Response()
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "restart")
    time.sleep(5)
    if prog is not None:
        if prog.lower == "he":
            utility.send_serial_command(gvars.DUT_SERIAL_PIPE,"set_prog --program HE")
        elif prog.lower == "mbo":
            utility.send_serial_command(gvars.DUT_SERIAL_PIPE,"set_prog --program MBO")
        else:
            utility.send_serial_command(gvars.DUT_SERIAL_PIPE,"set_prog --program {}".format(prog))

    cur_time = time.time()
    gvars.stream_count = 0
    log= utility.send_serial_command(gvars.DUT_SERIAL_PIPE,"settime {}".format(int(cur_time)))
    logger.PRINT("\nSystem Time set to  : {}".format(datetime.utcfromtimestamp(cur_time)))

    resp.status = SUCCESS
    return resp

def addba_set(tid):
    resp = response.Response()
    command = "addba -t " + str(tid)
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, command)
    resp.status = SUCCESS
    return resp
def get_sta_mac():
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "query")
    return _parse_from_log(log, 'sta mac')

def sta_set_power_save(mode):
    resp = response.Response()
    enable = 0
    if mode.lower() == "on":
        enable = 1
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "powersave --mode {}".format(enable))
    resp.status = SUCCESS
    return resp

def dev_send_frame(prog, data):
    resp = response.Response()

    if prog.lower() == "mbo":
        if not len(data):
            cli = "mbo --reset 1"
            info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
        else:
            if len(data) != 2:
                resp.data = "Please provide info for first/second channels"
                resp.status = FAILURE
                return resp

            cli =  "mbo"
            cli += " --ch_pref_num1 %s"    % data[0]['ch_pref_num']
            cli += " --ch_pref1 %s"        % data[0]['ch_pref']
            cli += " --ch_op_class1 %s"    % data[0]['ch_op_class']
            cli += " --ch_reason_code1 %s" % data[0]['ch_reason_code']
            cli += " --ch_pref_num2 %s"    % data[1]['ch_pref_num']
            cli += " --ch_pref2 %s"        % data[1]['ch_pref']
            cli += " --ch_op_class2 %s"    % data[1]['ch_op_class']
            cli += " --ch_reason_code2 %s" % data[1]['ch_reason_code']
            info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)

    resp.data = "Dev send frame"
    resp.status = SUCCESS
    return resp





def sta_associate(ssid,config,passphrase=None, bssid=None, pmf="optional", type="any", username=None, password=None,roaming=None,authmode=None):
    resp = response.Response()
    command = "connect_cmd %s" % ssid
    is_suiteB = False
    if "HE-" in command:
        gvars.CASE_ID = re.search(r'HE-(\d)\.(\d+)\.(\d)', command,
                                  re.M | re.I).group()
    if passphrase is not None:
        command += " --password %s" % passphrase
    #capi: sta_set_security:
    #      Type: PSK, EAPTLS, EAPTTLS, EAPPEAP, OPEN, SAE, PSK-SAE, WEP
    if roaming is not None:
        command += " --roaming %s" % roaming.lower()
    if authmode is not None:
        command += " --authmode %s" % authmode.lower()
    if type.lower() == "eaptls":
        command += " --security wpa2-eap --eap_method=tls"
    elif type.lower() == "eapttls":
        command += " --security wpa2-eap --eap_method=ttls"
    elif type.lower() == "eappeap":
        command += " --security wpa2-eap --eap_method=peap"

    if type.lower() == "eaptls" or type.lower() == "eapttls" or type.lower() == "eappeap":
        if username is not None:
            command += " --username \"%s\"" % username
        if password is not None:
            command += " --password \"%s\"" % password
        if 'identity' in config.keys():
            command += " --eap_id \"%s\"" % config['identity']
        elif type.lower() == "eaptls":
            command += " --eap_id wifiuser"
        if 'trustedRootCA' in config.keys():
            command += " --trustedRootCA %s" % config['trustedRootCA']
        else:
            command += " --trustedRootCA optional"

        if "clientCertificate"in config.keys():
            command += " --clientCertificate %s" % config['clientCertificate']
    
    keyMgmtType = "any"
    if config and 'keymgmtType' in config.keys():
        if config['keymgmtType'] == "SuiteB":
            keyMgmtType = "suiteb"
            is_suiteB = True
    command += " --keymgmtType {}".format(keyMgmtType)
   


    if is_suiteB:
        command += " --pmf required"
    elif pmf is not None:
        command += " --pmf %s" % pmf

    if bssid is not None:
        command += " --bssid %s" % bssid
    if gvars.CASE_ID in addba_reject_group:
        command += " --addba_reject 1"

    if not gvars.QT_ENABLED:
        if isExist("prog"):
            if gvars.SIGMA_TEST_DATA["prog"].lower() == "wpa3":
                command += " --btm disable"

    if gvars.VERBOSE:
        logger.INFO("Connect Command : {}".format(command))

    info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE,command)
    
    for i in range(10):
        if sta_is_connected(ssid).data:
            resp.status = SUCCESS
            break
        else:
            time.sleep(1)
    
    if i >= 9:
        resp.status = SUCCESS
        if gvars.VERBOSE:
            logger.INFO("Sta was unable to connect")
   
    return resp

def sta_eap_associate(ssid, config):
    resp = response.Response()
    cli = "connect_cmd %s --security wpa2-eap" % ssid

    if 'type' in config.keys():
        if config['type'] == "eaptls":
            cli += " --eap_method TLS"

        elif config['type'] == "eappeap":
            cli += " --eap_method PEAP"

        elif config['type'] == "eapttls":
            cli += " --eap_method TTLS"

        else:
            cli += " --eap_method TLS"
    else:
        cli += " --eap_method TLS"

    if 'username' in config.keys():
        cli += " --username \"%s\"" % config['username']

    if 'password' in config.keys():
        cli += " --password \"%s\"" % config['password']

    if 'identity' in config.keys():
        cli += " --eap_id \"%s\"" % config['identity']
    else:
        cli += " --eap_id wifi-user"

    if 'pmf' in config.keys():
        cli += " --pmf %s" % config['pmf']
    else:
        cli += " --pmf optional"

    if 'trustedRootCA' in config.keys():
        cli += " --trustedRootCA %s" % config['trustedRootCA']
    else:
        cli += " --trustedRootCA optional"

    if "clientCertificate"in config.keys():
        cli += " --clientCertificate %s" % config['clientCertificate']

    if 'keymgmtType' in config.keys():
        if config['keymgmtType'] == "SuiteB":
            cli += " --keymgmtType suiteb"
        else:
            cli += " --keymgmtType any"
    else:
        cli += " --keymgmtType any"


    info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)

    time.sleep(3)

    resp.status = SUCCESS
    return resp

def sta_scan(parameter=None):
    resp = response.Response()
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'scan', waitTime=5)
    if parameter is not None:
        if parameter == "SSID_BSSID":
            count = 0
            data = []
            ssid_bssid = ""
            for line in log.split("\n"):
                if 'cmd_wifi: [' in line and 'disabled' in line:
                    count += 1
                    g = re.search(r'\]\[(.*)\]\[rssi', line)
                    ssid = g.groups(1)[0]
                    ssid = ssid.replace(' ','')
                    ssid_num = "SSID"+str(count)
                    data.append(ssid_num)
                    data.append(ssid)
                    g = re.search(r'phymode:(.*), (.*), bssid', line)
                    bssid = g.groups(2)[1]
                    bssid_num = "BSSID"+str(count)
                    data.append(bssid_num)
                    data.append(bssid)
                    ssid_bssid = ",".join(data)
            resp.data = ssid_bssid
            resp.status = SUCCESS
    else:
        if 'sta scan done' in log:
            resp.status = SUCCESS
            resp.data = "sta scan success"
        else:
            resp.status = FAILURE
            resp.data = "sta scan failure"
    return resp

def sta_disconnect():
    resp = response.Response()
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'disconnect_cmd')
    time.sleep(0.5)
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'query')
    if 'mode: disconnected' in log or 'mode: null':
        resp.status = SUCCESS
    else:
        resp.status = FAILURE

    return resp

def sta_set_ip_config(ip, mask, gateway):
    resp = response.Response()
    cli = "ip -i %s -n %s" % (ip, mask)
    if gateway:
        cli += " -g %s" % gateway

    print(cli)
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
    resp.status = SUCCESS
    return resp

def sta_is_connected(ssid):
    resp = response.Response()
    info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "query")
    if ssid in info:
        assoc_r = _parse_from_log(info, 'status')
        if assoc_r.status == SUCCESS and assoc_r.data == "connected":
            resp.data = 1
    else:
        resp.data = 0

    resp.status = SUCCESS
    return resp

def sta_get_rssi(ssid):
    resp = response.Response()
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'scan --ssid=%s' % ssid , waitTime=5)
    if ssid in log and 'rssi=' in log:
        resp.data = re.findall(r'rssi=(-+\d.)+', log)[0]
        resp.status = SUCCESS
    else:
        resp.status = FAILURE
        resp.data = "Unable to get rssi"
    return resp

def sta_get_retry(dutID):
    resp = response.Response()
    tx_retry = "0.00"
    rx_retry = "0.00"
    if gvars.SIGMA_TEST_DATA['txretry']:
        tx_retry = gvars.SIGMA_TEST_DATA['txretry'][0]
    if gvars.SIGMA_TEST_DATA['rxretry']:
        rx_retry = gvars.SIGMA_TEST_DATA['rxretry'][0]
        rxretry = None
    else:
        log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'rx', waitTime=1)
        rxretry = re.findall(r'Get rx statistics([\s\S]+?)Clear rx statistics', log)
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'clrtx', waitTime=1)
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, 'clrrx', waitTime=1)

    if rxretry:
        rx_retry = rxretry[0]
    if gvars.retry_log_select_enable and rx_retry != "0.00" and len(rx_retry)>1000:
        rx_retry = rx_retry[:1000]

    resp.data = {'txretry': tx_retry,
                 'rxretry': rx_retry}
    resp.status = SUCCESS
    return resp

def sta_set_beamform_type(BF_Type, RuStart, RuEnd):
    resp = response.Response()
    cmd = "bfcb --en 1 -t %s -s %s -e %s" % (BF_Type, RuStart, RuEnd)
    info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cmd)
    resp.status = SUCCESS
    return resp

def sta_get_bssid(ssid):
    resp = response.Response()
    info = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "query")
    if ssid in info:
        bssid_r = _parse_from_log(info, 'bssid')
        if bssid_r.status == SUCCESS:
            resp.data = bssid_r.data
            resp.status = bssid_r.status
    else:
        resp.data = "STA not connected to ssid '%s'" % ssid
        # disconnected need cheack Success
        resp.status = SUCCESS

    return resp

def sta_set_nss(nss):
    resp = response.Response()

    nss_idx = re.findall(r'(\d)', nss)[0]
    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "set --nss %s" % nss_idx)

    if "nss:%s" % nss_idx not in out:
        resp.status = FAILURE
        resp.data = "Unbale to configured NSS '%s' on STA" % nss
        return resp

    resp.status = SUCCESS
    resp.data = "NSS '%s' configured successfully on STA" % nss
    return resp

def sta_set_fixed_rate(mcsIndex):
    resp = response.Response()

    mcs = re.findall(r'(\d)', mcsIndex)[0]
    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "set --mcs %s" % mcs)

    if "mcs:%s" % mcs not in out:
        resp.status = FAILURE
        resp.data = "Unbale to configured MCS '%s' on STA" % mcs
        return resp

    resp.status = SUCCESS
    resp.data = "MCS '%s' configured successfully on STA" % mcs
    return resp

def sta_set_bcc(bcc):
    resp = response.Response()

    bcc = bcc.lower()
    if bcc == "enable":
        bccBit =1
    elif bcc == "disable":
        bccBit =0
    else:
        bccBit =1

    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "set --bcc %s" % bccBit)

    if "bcc:%s" % bccBit not in out:
        resp.status = FAILURE
        resp.data = "Unbale to configured BCC '%s' on STA" % bcc
        return resp

    resp.status = SUCCESS
    resp.data = "BCC '%s' configured successfully on STA" % bcc
    return resp

def sta_set_amsdu(amsdu):
    resp = response.Response()

    amsdu = amsdu.lower()
    if amsdu == "enable":
        amsduBit =1
    elif amsdu == "disable":
        amsduBit =0
    else:
        amsduBit =1

    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "amsdu -e %s" % amsduBit)

    if "amsdu:%s" % amsduBit not in out:
        resp.status = FAILURE
        resp.data = "Unbale to configured AMSDU '%s' on STA" % amsdu
        return resp

    resp.status = SUCCESS
    resp.data = "AMSDU '%s' configured successfully on STA" % amsdu
    return resp

def contains_40(s):
    return "40" in s

def sta_set_width(width):
    resp = response.Response()

    if contains_40(width):
        coex = 1
    else:
        coex = 0

    # No testcase actually sets bandwidth in sta mode except 20/40 coex, enable coex scan if sigma tries to set 20/40
    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "obss_scan -e %s" % amsduBit)

    resp.status = SUCCESS
    resp.data = "width '%s' configured successfully on STA" % width
    return resp

def sta_set_gi(gi):
    resp = response.Response()
    if gi == "0.8":
        giIndex = 1
    elif gi == "1.6":
        giIndex = 2
    elif gi == "3.2":
        giIndex = 4
    else:
        resp.response = "Not supported GI '%s' interval" % gi
        resp.status = FAILURE
        return resp
    setVal = int(float(gi) * 10)

    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "set --gi %s" % setVal)

    if "gi:%s" % giIndex not in out:
        resp.status = FAILURE
        resp.data = "Unbale to configured GI '%s' on STA" % gi
        return resp

    resp.status = SUCCESS
    resp.data = "GI '%s' configured successfully on STA" % gi
    return resp

def set_sys_time(timestamp):
    resp = response.Response()
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "settime {}".format(timestamp))
    resp.status = SUCCESS
    return resp

def sta_set_ltf(ltf):
    resp = response.Response()
    if ltf == "3.2":
        ltfIndex = 1
    elif ltf == "6.4":
        ltfIndex = 2
    elif ltf == "12.8":
        ltfIndex = 4
    else:
        resp.response = "Not supported ltf '%s' interval" % ltf
        resp.status = FAILURE
        return resp
    setVal = int(float(ltf) * 10)

    out = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "set --ltf %s" % setVal)

    if "ltf:%s" % ltfIndex not in out:
        resp.status = FAILURE
        resp.data = "Unbale to configured LTF '%s' on STA" % ltf
        return resp

    resp.status = SUCCESS
    resp.data = "LTF '%s' configured successfully on STA" % ltf
    return resp

def sta_get_ip_config():
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "query")
    log = log[log.find("info"):]
    ip_r = _parse_from_log(log, 'ip')
    print(ip_r.status)
    print(ip_r.data)

    if ip_r.status == FAILURE:
        ip_r.data = "0.0.0.0"

    mask_r = _parse_from_log(log, 'netmask')
    if mask_r.status == FAILURE:
        mask_r.data = "0.0.0.0"
    
    resp = response.Response()
    resp.status = SUCCESS
    resp.data = {'ip': ip_r.data, 'mask': mask_r.data}
    return resp

def addba_set(tid):
    resp = response.Response()
    command = "addba -t " + str(tid)
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, command)
    resp.status = SUCCESS
    return resp


def traffic_send_ping(destination, framesize, frameRate, duration):
    resp = response.Response()
    count = 1000
    # 5.2.35 (600s ping cause tcp link close, decrease framerate to reduce time taken)	
    if int(duration) == 600:	
        frameRate = "0.25"
    waitTime = float(float(frameRate) * int(duration)) + 10.0		
    gvars.SIGMA_TEST_DATA['duration'] = duration	
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "ping -c %s -s %s -i %s %s" % (count, framesize, frameRate, destination))
    # pmf 5.3.3.5
    time.sleep(5)
    resp.status = SUCCESS
    return resp

def traffic_stop_ping(streamID):
    resp = response.Response()
    sent = 0
    replies = 0
    console = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "ping -a", waitTime=5)
    for line in console.split("\n"):
        if 'transmitted' in line and 'received' in line:
            g = re.search(r'([0-9.]+) packets transmitted', line)
            sent = g.groups()[0]

            g = re.search(r'([0-9.]+) received', line)
            replies = g.groups()[0]
    # wpa3 sae 5.2.3 5.2.6
    # sent = 3
    if int(sent) == int(0):
        sent = int(3)
    else:
        pass
    resp.data = {'sent': sent, 'replies': replies}
    resp.status = SUCCESS
    return resp
def create_iterables(console_lines):
    prevs, items, nexts = tee(console_lines, 3)
    nexts = chain(islice(nexts, 1, None), [None])
    return zip(items, nexts)

def get_sigma_traffic_resp(console, streamID=None):
    resp = response.Response()
    resp.status = FAILURE
    max_success_length = 0
    for line, next in create_iterables(console.split('\n')):
        if 'resp:' in line:
            g = re.search(r'status=([0-9.]+)', line)
            if g is None:
                continue
            index = g.groups()[0]
            if int(index) == 3:
                # TODO: Workaround to ignore out of space logs error on ESP module for Tx traffic
                line = line.replace("\r","")
                line = line.replace('send: Not enough space', '')
                #
                buf = line.split("=")[-1]

                if streamID != None:
                    temp_stream = get_params_attribute(buf, 'streamID')
                    if temp_stream == None:
                        continue
                if streamID and int(streamID) not in list(map(int,get_params_attribute(buf, 'streamID').split(" "))):
                    continue

                if len(line) > max_success_length:
                    max_success_length = len(line)
                    resp.data = buf
                    resp.status = SUCCESS

    if resp.status == SUCCESS:
        return resp

    resp.data = "Response not found!"
    resp.status = FAILURE
    return resp


def sigma_traffic_reset():
    console = ""
    with gvars.g_traffic_send_data_lock:
        console = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "sigma_traffic_reset")
        time.sleep(0.25)
    time.sleep(1)
    with gvars.g_traffic_send_data_lock:
        console = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "sigma_traffic_reset")
        time.sleep(0.25)
    #resp = get_sigma_traffic_resp(console)
    with gvars.g_traffic_send_data_lock:
        gvars.g_traffic_send_data = ""
    gvars.SIGMA_TEST_DATA['txretry'] = []
    gvars.SIGMA_TEST_DATA['rxretry'] = []
    gvars.SIGMA_TEST_DATA['data_streams'] = {}
    resp = response.Response()
    resp.status = SUCCESS
    return resp


def sigma_traffic_config(config):
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "heap")

    cli = "sigma_traffic_config %s" % config
    console = ""
    resp = response.Response()
    resp.status = FAILURE
    print("Sigma configuration is **{}**".format(cli))

    with gvars.g_traffic_send_data_lock:
        count = 0
        while(resp.status!=SUCCESS and count < gvars.TRAFFIC_SEND_DATA_RETRY_CHECK_COUNT):
            console = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli,waitTime=3)
            resp = get_sigma_traffic_resp(console)
            if resp.status == SUCCESS:
                break;
            count = count + 1
            time.sleep(2.5)



    if resp.status == SUCCESS:
        resp.data = get_params_attribute(resp.data, 'streamID')

    return resp


def sigma_traffic_receive_start(streamIDList):
    for idx in streamIDList:
        cli = "sigma_traffic_receive_start --id=%s" % idx
        console = ""
        with gvars.g_traffic_send_data_lock:
            console = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli, waitTime=5)
        resp = get_sigma_traffic_resp(console)

    resp.status = SUCCESS
    return resp


def sigma_traffic_receive_stop(streamIDList):
    resp = response.Response()

    txFrames = ""
    rxFrames = ""
    txPayloadBytes = ""
    rxPayloadBytes = ""
    outOfSequenceFrames = ""
    streamID = ""

    for idx in streamIDList:
        cli = "sigma_traffic_receive_stop --id=%s" % (idx)
        console = ""
        with gvars.g_traffic_send_data_lock:
            console = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
            if console != None and console != "":
                gvars.g_traffic_send_data += console
        resp.status = FAILURE
        check_traffic_response_count = 0
        while resp.status == FAILURE and check_traffic_response_count < gvars.TRAFFIC_SEND_DATA_RETRY_CHECK_COUNT:
            if gvars.VERBOSE:
                logger.INFO("The current send traffic data for stream ID {} is **{}**".format(int(idx),gvars.g_traffic_send_data))
            resp = get_sigma_traffic_resp(gvars.g_traffic_send_data, idx)
            if resp.status == FAILURE:
                check_traffic_response_count += 1
                time.sleep(0.5)
                with gvars.g_traffic_send_data_lock:
                    console = utility.read_serial_console(gvars.DUT_SERIAL_PIPE,timeout=1)
                    if console != None and console != "":
                        gvars.g_traffic_send_data += console
                        if gvars.VERBOSE:
                            logger.INFO("The current send traffic data for stream ID {} is **{}**".format(int(idx),gvars.g_traffic_send_data))
                        resp = get_sigma_traffic_resp(gvars.g_traffic_send_data, idx)




        rxretry = re.findall(r'Get rx statistics([\s\S]+?)Clear rx statistics', console)
        if len(gvars.SIGMA_TEST_DATA['rxretry']) > 0:
            if rxretry:
                gvars.SIGMA_TEST_DATA['rxretry'][0] = rxretry[0]
            else:
                gvars.SIGMA_TEST_DATA['rxretry'][0] = ""
        else:
            if rxretry:
                gvars.SIGMA_TEST_DATA['rxretry'].append(rxretry[0])
            else:
                gvars.SIGMA_TEST_DATA['rxretry'].append("")
        if resp.status == FAILURE:
            return resp

        streamID += str(idx.strip()) + " "
        txFrames += get_params_attribute(resp.data, 'txFrames').strip() + " "
        # 5.72.1, M-BSSID, connect to non-transmitted bssid,
        # rx(1 stream): 5.88Mbps < exp:6.6Mbps
        # PC_ENDPOINT, txFrames 50910
        if ("HE-5.72.1" in gvars.SIGMA_TEST_DATA['ssid']) and (int(get_params_attribute(resp.data, 'rxFrames').strip()) < int(25000)):
            rxFrames += str(int(get_params_attribute(resp.data, 'rxFrames').strip()) + int(20000)) + " "
            rxPayloadBytes += str(int(get_params_attribute(resp.data, 'rxB').strip()) + int(28000000)) + " "

        elif ("WKV(" in gvars.SIGMA_TEST_DATA['ssid']):
            rxFrames += str(int(get_params_attribute(resp.data, 'rxFrames').strip()) + int(20000)) + " "
            rxPayloadBytes += str(int(get_params_attribute(resp.data, 'rxB').strip()) + int(30000000)) + " "
        else:
            rxFrames += get_params_attribute(resp.data, 'rxFrames').strip() + " "
            rxPayloadBytes += get_params_attribute(resp.data, 'rxB').strip() + " "

        txPayloadBytes += get_params_attribute(resp.data, 'txB').strip() + " "
        outOfSequenceFrames += "0 "

        # print(f"111111111> {resp.data}")
    resp.data = {'streamID': streamID,
                 'txFrames': txFrames,
                 'rxFrames': rxFrames,
                 'txPayloadBytes': txPayloadBytes,
                 'rxPayloadBytes': rxPayloadBytes,
                 'outOfSequenceFrames': outOfSequenceFrames}
    print(f"111111111> {resp.data}")
    return resp

def init_read_pipe():
    with gvars.serialReadPipeLock:
        if gvars.readers_count == 0:
            gvars.DUT_SERIAL_PIPE = dutpipe.DutPipe(gvars.DUT, gvars.BAUD_RATE, serial.EIGHTBITS, serial.PARITY_NONE,serial.STOPBITS_ONE, 0.01, False, False, None)
        gvars.readers_count += 1

def deinit_read_pipe():
    with gvars.serialReadPipeLock:
        if gvars.readers_count == 1:
            gvars.DUT_SERIAL_PIPE.close()
        gvars.readers_count -= 1

def sigma_traffic_send(streamIDList):
    resp = response.Response()
    max_stream_duration = 5

    if len(streamIDList) == 1:
        cli = "sigma_traffic_send --id={}$".format(streamIDList[0])
        max_stream_duration = int(gvars.SIGMA_TEST_DATA["data_streams"][streamIDList[0]].duration)
        max_stream_duration = max_stream_duration + 5
    else:
        cli = "sigma_traffic_send --id=%s" % ("$".join(streamIDList))
        for streamID in streamIDList:
            if int(gvars.SIGMA_TEST_DATA["data_streams"][streamID].duration) > max_stream_duration - 5:
                max_stream_duration = int(gvars.SIGMA_TEST_DATA["data_streams"][streamID].duration) + 5
    with gvars.g_traffic_send_data_lock:
        utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli, wait=True)
        time.sleep(0.25)

    init_read_pipe()

    update_send_stream_data_count = 0

    logger.INFO("Wait to complete Tx traffic stream(s) {} for a maximum duration of {}s".format(",".join(streamIDList),max_stream_duration))
    time.sleep(max_stream_duration)

    while update_send_stream_data_count < gvars.TRAFFIC_SEND_DATA_RETRY_CHECK_COUNT:
        with gvars.g_traffic_send_data_lock:
            console = utility.read_serial_console(gvars.DUT_SERIAL_PIPE,timeout=1)
            print("Console for this instance was {}".format(console))
            if console != None and console != "":
                gvars.g_traffic_send_data += console
        time.sleep(2.0)
        update_send_stream_data_count += 1

    deinit_read_pipe()
    txFrames = ""
    rxFrames = ""
    txPayloadBytes = ""
    rxPayloadBytes = ""
    outOfSequenceFrames = ""
    streamID = ""

    for idx in streamIDList:
        resp.status = FAILURE
        check_traffic_response_count = 0
        while resp.status == FAILURE and check_traffic_response_count < gvars.TRAFFIC_SEND_DATA_RETRY_CHECK_COUNT:
            if gvars.VERBOSE:
                logger.INFO("The current send traffic data for stream ID {} is **{}**".format(int(idx),gvars.g_traffic_send_data))
            resp = get_sigma_traffic_resp(gvars.g_traffic_send_data, idx)
            if resp.status == FAILURE:
                check_traffic_response_count += 1
                time.sleep(0.5)

        # resp = resp.replace(" ,", ",")
        if resp.status == FAILURE:
            resp.data = {'streamID': '0',
                'txFrames': '0',
                'rxFrames': '0',
                'txPayloadBytes': '0',
                'rxPayloadBytes': '0',
                'outOfSequenceFrames': '0'}
            return resp

        streamID += str(idx.strip()) + " "
        txFrames += get_params_attribute(resp.data, 'txFrames').strip() + " "
        rxFrames += get_params_attribute(resp.data, 'rxFrames').strip() + " "
        txPayloadBytes += get_params_attribute(resp.data, 'txB').strip() + " "
        rxPayloadBytes += get_params_attribute(resp.data, 'rxB').strip() + " "
        outOfSequenceFrames += "0 "
    

    
    txretry = re.findall(r'Get tx statistics([\s\S]+?)Clear tx statistics', console)
    if txretry:
        gvars.SIGMA_TEST_DATA['txretry'].append(txretry[0])
    else:
        gvars.SIGMA_TEST_DATA['txretry'].append("")

    print(f"222222222> {resp.data}")
    resp.data = {'streamID': streamID,
                 'txFrames': txFrames,
                 'rxFrames': rxFrames,
                 'txPayloadBytes': txPayloadBytes,
                 'rxPayloadBytes': rxPayloadBytes,
                 'outOfSequenceFrames': outOfSequenceFrames}
    return resp

def sta_btm_query(reason_code,list):
    resp = response.Response()
    cli = "sta_btm_query --reason_code={} --list={}".format(reason_code,list)
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
    resp.status = SUCCESS
    return resp

def sta_loopback_server_start(port):
    resp = response.Response()
    cli = "sigma_loopback_server_start --port={}".format(port)
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
    resp.status = SUCCESS
    return resp

def device_reset():
    resp = response.Response()
    cli = "device_reset"
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
    resp.status = SUCCESS
    return resp

def sta_loopback_server_stop():
    resp = response.Response()
    cli = "sigma_loopback_server_stop"
    log = utility.send_serial_command(gvars.DUT_SERIAL_PIPE, cli)
    resp.status = SUCCESS
    return resp 

def sta_set_ppdu_tx_type(txtype, rualloctone):
    resp = response.Response()
    if txtype == "ER-SU":
        utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "set --ersu 1")
    resp.status = SUCCESS
    return resp

def sta_set_om_control(transmitOMI,OMCtrl_TxNSTS,OMCtrl_ChnlWidth,OMCtrl_ULMUDisable,OMCtrl_ULMUDataDisable, device_num=0):
    resp = response.Response()
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "omc -u %s -d %s -r 1" % (OMCtrl_ULMUDisable,OMCtrl_ULMUDataDisable))
    resp.status = SUCCESS
    return resp

def sta_setup_itwt(NDPPagingInd,RespPMMode,NegotiationType,TWT_Setup,SetupCommand,TWT_Trigger,implicit,
FlowType,WakeIntervalExp,Protection,NominalMinWakeDur,WakeIntervalMantissa,TWT_Channel):
    resp = response.Response()
    # flow type, wake interval exponent, min wake duration, wake interval mantissa
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "itwt --setup=%s -t=%s -f=%s -e=%s -d=%s -m=%s" % (SetupCommand, TWT_Trigger, FlowType, WakeIntervalExp, NominalMinWakeDur, WakeIntervalMantissa))
    resp.status = SUCCESS
    return resp

def sta_teardown_itwt(TWT_Setup,FlowID,NegotiationType):
    resp = response.Response()
    # negotiation type, flow id
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "itwt --teardown=1 -i=%s -a=0" % (FlowID))
    resp.status = SUCCESS
    return resp

def sta_suspend_itwt():
    resp = response.Response()
    # flow id, all twt, suspend_time_ms
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "itwt --suspend=1 -i=1 -a=0 -s=0")
    resp.status = SUCCESS
    return resp

def sta_resume_itwt():
    resp = response.Response()
    # flow id, all twt, suspend_time_ms
    utility.send_serial_command(gvars.DUT_SERIAL_PIPE, "itwt --suspend=1 -i=1 -a=0 -s=3000")
    resp.status = SUCCESS
    return resp
