#!/usr/bin/env python3
""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """
import time
import socket
import threading
import argparse
import serial
import signal
import sys
import os
import re

from common import gvars, logger, dutpipe, utility
from capi import capi
from quicktrack import quicktrack

def keyboardInterruptHandler(signal, frame):
    logger.ERROR("KeyboardInterrupt (ID: {}) has been caught. Cleaning up...".format(signal))
    gvars.DUT_SERIAL_PIPE.close()
    gvars.SERVER.close()
    exit(0)

connect_change = False
class Dut_Send(threading.Thread):
    def __init__(self,client_socket,capi_call):
        threading.Thread.__init__(self)
        self.client_socket = client_socket
        self.capi_call = capi_call

    def run(self):
        self.client_socket.send(gvars.CAPI_RESP_RUNNING.encode() + b'\r\n')
        response = capi.process_ucc_request(self.capi_call)
        self.client_socket.send(response.encode() + b'\r\n')

def handle_ucc_connection(client_socket):
    """ Handles UCC machine connection requests

    Parameters:
        client socket (socket): Socket client opened by UCC

    Returns:
        None:

    """
    while True:
        try:
            ucc_capi_call = client_socket.recv(gvars.BUFFER).decode()
        except:
            logger.INFO('Connection reset from UCC')
            client_socket.close()
            break

        if not ucc_capi_call.strip():
            client_socket.close()
            break

        dut_thread = Dut_Send(client_socket, ucc_capi_call)
        dut_thread.start()

def handle_qt_connection(client_socket):
    """ Handles QT machine connection requests

    Parameters:
        client socket (socket): Socket client opened by QT

    Returns:
        None:

    """
    while True:
        try:
            qt_call = client_socket.recv(gvars.BUFFER).decode()
        except:
            logger.INFO('Connection reset from QT')
            client_socket.close()
            break

        if not qt_call.strip():
            client_socket.close()
            break

        response = quicktrack.process_qt_request(qt_call)
        logger.INFO("Sending Response : {}".format(response))
        client_socket.send(response.encode())
        break

class Runclient(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
    def run(self):
        global connect_change
        while True:
            logger.INFO('ESP Sigma listening on {}:{}\n'.format(gvars.BIND_IP, gvars.BIND_PORT))
            client_sock, address = gvars.SERVER.accept()
            gvars.client_socket = client_sock
            logger.INFO('Accepted connection from {}:{}'.format(address[0], address[1]))
            gvars.threadLock.acquire()
            connect_change = True
            gvars.threadLock.release()
            time.sleep(1)
            handle_ucc_connection(client_sock)

if __name__ == "__main__":
    signal.signal(signal.SIGINT, keyboardInterruptHandler)
    _time_str = time.strftime("%Y_%m_%d_%H:%M:%S", time.localtime())
    gvars.LOG_FILE_LOCATION = "logs/dut_{}.log".format(_time_str)
    sys.stdout = logger.Logger()
    parser = argparse.ArgumentParser(description='Espressif CAPI Sigma Agent Software')
    parser.add_argument('--dut', '-d', default="N/A", help='Espressif DUT')
    parser.add_argument('--program', '-p', default=gvars.PROGRAM, help='WFA Test programs (Ex. 11N, P2P, PMF, MBO etc.)')
    parser.add_argument('--verbose', '-v', action='store_true', help='Verbose Sigma logs')
    parser.add_argument('--quicktrack','-qt',action='store_true', help='Enable QuickTrack Compatibility')
    args = parser.parse_args()

    gvars.VERBOSE = args.verbose
    gvars.QT_ENABLED = args.quicktrack
    gvars.PROGRAM = args.program

    if args.dut is "N/A":
        cli = "ls /dev/ttyUSB*"
        log = utility.send_local_command(cli)
        if len(re.findall('/dev/ttyUSB',log)) == 1:
            gvars.DUT = log.strip()
            if gvars.VERBOSE:
                logger.PRINT("\nSet DUT port to : {}".format(gvars.DUT))

    utility.sync_ca_dut_prompts()

    if gvars.QT_ENABLED :
        if os.path.exists(gvars.SOCKET_BIND_LOCATION):
            os.remove(gvars.SOCKET_BIND_LOCATION)
        gvars.SERVER = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        gvars.SERVER.bind(gvars.SOCKET_BIND_LOCATION)
    else:
        gvars.SERVER = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        gvars.SERVER.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        gvars.SERVER.bind((gvars.BIND_IP, gvars.BIND_PORT))
        gvars.SERVER.listen(100)

    logger.PRINT("%s" % "*" * 75)

    if gvars.QT_ENABLED:
        logger.PRINT('Espressif Wi-Fi Alliance QuickTrack Test Tool Sigma DUT Agent\n')
    else:
        logger.PRINT('Espressif Wi-Fi Alliance Sigma DUT Agent\n')

    logger.PRINT("%-20s: %-20s" % ("WFA PROGRAM", gvars.PROGRAM))
    logger.PRINT("%-20s: %-20s" % ("DUT PORT", gvars.DUT))
    logger.PRINT("%-20s: %-20s" % ("IDF Version", gvars.IDF_VERSION))
    logger.PRINT("%-20s: %-20s" % ("SIGMA CA", gvars.SW_VERSION))

    logger.PRINT("")
    logger.PRINT("%s" % "*" * 75)

    if gvars.QT_ENABLED:
        while True:
            gvars.SERVER.listen(100)
            logger.INFO('ESP Sigma listening at {}\n'.format(gvars.SOCKET_BIND_LOCATION))
            client_sock, address = gvars.SERVER.accept()
            logger.INFO('Accepted connection from {}'.format(address))
            handle_qt_connection(client_sock)
            client_sock.close()
    else:
        try:
            gvars.threadLock = threading.Lock()
            #read_thread = ReadPort()
            #read_thread.start()
            run_client = Runclient()
            run_client.start()
        except Exception as ex:
            print(ex)
    gvars.DUT_SERIAL_PIPE.close()
