
#ifndef ESP_SIGMA_H
#define ESP_SIGMA_H

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <esp_netif.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <pthread.h>

#define CONFIG_WMM_POWER_SAVE CONFIG_ESPSIGMA_ENABLE_UAPSD

#define PRINTF_FORMAT(a,b)

#define SIGMA_TMPDIR "/tmp"

#define SIGMA_DUT_VER "(unknown)"

#define ETH_ALEN 6

#define ETH_P_ARP 0x0806

#define ARRAY_SIZE(x) (sizeof((x)) / (sizeof(((x)[0]))))

#define IPV6_ADDR_LEN 16

#define MAX_CMD_LEN 4096

#define MAX_PARAMS 100

#define MAX_SIGMA_STREAMS 6

#define MAX_SIGMA_STATS 100

struct sigma_dut;

typedef unsigned int u32;
typedef uint16_t u16;
typedef unsigned char u8;

#define WPA_GET_BE32(a) ((((u32) (a)[0]) << 24) | (((u32) (a)[1]) << 16) | \
             (((u32) (a)[2]) << 8) | ((u32) (a)[3]))
#define WPA_PUT_BE32(a, val)                    \
    do {                            \
        (a)[0] = (u8) ((((u32) (val)) >> 24) & 0xff);   \
        (a)[1] = (u8) ((((u32) (val)) >> 16) & 0xff);   \
        (a)[2] = (u8) ((((u32) (val)) >> 8) & 0xff);    \
        (a)[3] = (u8) (((u32) (val)) & 0xff);       \
    } while (0)

struct sigma_cmd {
    char *params[MAX_PARAMS];
    char *values[MAX_PARAMS];
    int count;
};

struct sigma_conn {
    int s;
    struct sockaddr_in addr;
    socklen_t addrlen;
    char buf[MAX_CMD_LEN + 5];
    int pos;
    int waiting_completion;
};

enum sigma_cmd_result {
    STATUS_SENT_ERROR = -3,
    ERROR_SEND_STATUS = -2,
    INVALID_SEND_STATUS = -1,
    STATUS_SENT = 0,
    SUCCESS_SEND_STATUS = 1
};

struct sigma_frame_stats {
    unsigned int seqnum;
    unsigned int local_sec;
    unsigned int local_usec;
    unsigned int remote_sec;
    unsigned int remote_usec;
};

struct sigma_stream {
    enum sigma_stream_profile {
        SIGMA_PROFILE_FILE_TRANSFER,
        SIGMA_PROFILE_MULTICAST,
        SIGMA_PROFILE_IPTV,
        SIGMA_PROFILE_TRANSACTION,
        SIGMA_PROFILE_START_SYNC,
        SIGMA_PROFILE_UAPSD
    } profile;
    int sender;
    struct in_addr dst;
    int dst_port;
    struct in_addr src;
    int src_port;
    int frame_rate;
    int duration;
    unsigned int payload_size;
    int start_delay;
    int max_cnt;
    enum sigma_traffic_class {
        SIGMA_TC_VOICE,
        SIGMA_TC_VIDEO,
        SIGMA_TC_BACKGROUND,
        SIGMA_TC_BEST_EFFORT
    } tc;
    int user_priority;
    int user_priority_set;
    int started;
    int no_timestamps;

    int sock;
    pthread_t thr;
    int stop;
    int ta_send_in_progress;
    int trans_proto;

    /* Statistics */
    int tx_act_frames; /*
                * Number of frames generated by the traffic
                * generator application. The name is defined in the
                * Sigma CAPI spec.
                */
    int tx_frames;
    int rx_frames;
    unsigned long long tx_payload_bytes;
    unsigned long long rx_payload_bytes;
    int out_of_seq_frames;
    struct sigma_frame_stats *stats;
    unsigned int num_stats;
    unsigned int stream_id;

    /* U-APSD */
    unsigned int sta_id;
    unsigned int rx_cookie;
    unsigned int uapsd_sta_tc;
    unsigned int uapsd_rx_state;
    unsigned int uapsd_tx_state;
    unsigned int tx_stop_cnt;
    unsigned int tx_hello_cnt;
    pthread_t uapsd_send_thr;
    pthread_cond_t tx_thr_cond;
    pthread_mutex_t tx_thr_mutex;
    int reset_rx;
    int num_retry;
    char ifname[8]; /* ifname from the command */
    struct sigma_dut *dut; /* for traffic agent thread to access context */
    /* console */
    char test_name[9]; /* test case name */
    int can_quit;
    int reset;
};

struct sigma_dut {
    int debug_level;
    int stdout_debug;
    struct sigma_cmd_handler *cmds;
    int response_sent;

    /* Traffic Agent */
    struct sigma_stream streams[MAX_SIGMA_STREAMS];
    int stream_id;
    int num_streams;
    pthread_t thr;

    unsigned int throughput_pktsize; /* If non-zero, override pktsize for
                      * throughput tests */
    int no_timestamps;

    enum sigma_program {
        PROGRAM_UNKNOWN = 0,
        PROGRAM_TDLS,
        PROGRAM_HS2,
        PROGRAM_HS2_R2,
        PROGRAM_WFD,
        PROGRAM_DISPLAYR2,
        PROGRAM_PMF,
        PROGRAM_WPS,
        PROGRAM_60GHZ,
        PROGRAM_HT,
        PROGRAM_VHT,
        PROGRAM_NAN,
        PROGRAM_LOC,
        PROGRAM_MBO,
        PROGRAM_IOTLP,
        PROGRAM_DPP,
        PROGRAM_OCE,
        PROGRAM_WPA3,
        PROGRAM_HE,
        PROGRAM_HS2_R3,
    } program;

    const char *summary_log;
    int iface_down_on_reset;
    int write_stats; /* traffic stream e2e*.txt files */

    const char *vendor_name; /* device_get_info vendor override */
    const char *model_name; /* device_get_info model override */
    const char *version_name; /* device_get_info version override */
    const char *log_file_dir; /* Directory to generate log file */
    FILE *log_file_fd; /* Pointer to log file */
    struct sigma_cmd c;
};

enum sigma_dut_print_level {
    DUT_MSG_DEBUG, DUT_MSG_INFO, DUT_MSG_ERROR
};

void sigma_dut_print(struct sigma_dut *dut, int level, const char *fmt, ...)
PRINTF_FORMAT(3, 4);

void sigma_dut_dev_print(const char *fmt, ...)
PRINTF_FORMAT(3, 4);

void sigma_dut_summary(struct sigma_dut *dut, const char *fmt, ...)
PRINTF_FORMAT(2, 3);

enum sigma_status {
    SIGMA_RUNNING, SIGMA_INVALID, SIGMA_ERROR, SIGMA_COMPLETE
};

enum sigma_traffic_cmd {
    SIGMA_TRAFFIC_CONFIG, SIGMA_TRAFFIC_RECEIVE_START, SIGMA_TRAFFIC_RECEIVE_STOP, SIGMA_TRAFFIC_SEND, SIGMA_TRAFFIC_RESET
};

void send_resp(struct sigma_dut *dut,
               enum sigma_status status, const char *buf);

const char * get_params(struct sigma_cmd *cmd, const char *name);
const char * get_param_indexed(struct sigma_cmd *cmd, const char *name,
                               int index);

#if CONFIG_WMM_POWER_SAVE
void receive_uapsd(struct sigma_stream *s);

void send_uapsd_console(struct sigma_stream *s);
#endif /*CONFIG_WMM_POWER_SAVE*/

enum sigma_cmd_result cmd_traffic_agent_config(struct sigma_dut *dut,
                                               struct sigma_cmd *cmd);

enum sigma_cmd_result cmd_traffic_agent_receive_start(struct sigma_dut *dut,
                                                      struct sigma_cmd *cmd);

enum sigma_cmd_result cmd_traffic_agent_receive_stop(struct sigma_dut *dut,
                                                     struct sigma_cmd *cmd);

enum sigma_cmd_result cmd_traffic_agent_reset(struct sigma_dut *dut,
                                              struct sigma_cmd *cmd);

enum sigma_cmd_result cmd_traffic_agent_send(struct sigma_dut *dut,
                                             struct sigma_cmd *cmd);

enum sigma_cmd_result cmd_traffic_agent_version(struct sigma_dut *dut,
                                                struct sigma_cmd *cmd);

void process_sigma_traffic_cmd(struct sigma_dut *, char *, int);

#endif /* ESP_SIGMA_H */
